"""
$(SIGNATURES)
using DataFrames: AbstractAggregate
Combine two UnfoldDesignmatrices. This allows combination of multiple events.

This also allows to define events with different lengths.

Not supported for models without timebasis, as it is not needed there (one can simply run multiple models)
# Examples
```julia-repl
julia>  basisfunction1 = firbasis(τ=(0,1),sfreq = 10,name="basis1")
julia>  basisfunction2 = firbasis(τ=(0,0.5),sfreq = 10,name="basis2")
julia>  Xdc1          = designmatrix(UnfoldLinearModelContinuousTime(Dict(Any=>(@formula 0~1,basisfunction1)),tbl_1)
julia>  Xdc2          = designmatrix(UnfoldLinearModelContinuousTime(Dict(Any=>(@formula 0~1,basisfunction2)),tbl_2)
julia>  Xdc = Xdc1+Xdc2 
```

"""
function combineDesignmatrices(X1::DesignMatrix, X2::DesignMatrix)

    # the reason for the assertion is simply that I found it too difficult to concatenate the formulas down below ;) should be easy to implement hough
    @assert !(isa(X1.formulas,AbstractArray) && isa(X2.formulas,AbstractArray)) "it is currently not possible to combine desigmatrices from two already concatenated designs - please concatenate one after the other"

    X1 = deepcopy(X1)
    X2 = deepcopy(X2)
    Xs1 = get_Xs(X1)
    Xs2 = get_Xs(X2)

    sX1 = size(Xs1, 1)
    sX2 = size(Xs2, 1)

    # append 0 to the shorter designmat
    if sX1 < sX2
        Xs1 = SparseMatrixCSC(sX2, Xs1.n, Xs1.colptr, Xs1.rowval, Xs1.nzval)
    elseif sX2 < sX1
        Xs2 = SparseMatrixCSC(sX1, Xs2.n, Xs2.colptr, Xs2.rowval, Xs2.nzval)
    end
    if typeof(X1.Xs) <: Matrix
        # mass univariate case, simply put in a vector
        if typeof(Xs1) <:Vector
            Xcomb = [Xs1...,Xs2]
        else
            Xcomb = [Xs1,Xs2]
        end
    else
        Xcomb = hcat(Xs1, Xs2)
    end

    #if !(Float64(X1.formulas.rhs.basisfunction.times.step) ≈ Float64(X2.formulas.rhs.basisfunction.times.step))
    #        @warn("Concatenating formulas with different sampling rates. Be sure that this is what you want.")
    #end
    
    if typeof(X1.Xs) <: Tuple
        # we have random effects                
        # combine REMats in single-eventtpe formulas ala y ~ (1|x) + (a|x)
        Xs1 = MixedModels._amalgamate([X1.Xs[2:end]...], Float64)
        Xs2 = MixedModels._amalgamate([X2.Xs[2:end]...], Float64)

        Xcomb = (Xcomb, Xs1..., Xs2...)

        # Next we make the ranefs all equal size
        equalizeReMatLengths!(Xcomb[2:end])

        # check if ranefs can be amalgamated. If this fails, then MixedModels tried to amalgamate over different eventtypes and we should throw the warning
        # if it success, we have to check if the size before and after is identical. If it is not, it tried to amalgamize over different eventtypes which were of the same length

        try
            reterms = MixedModels._amalgamate([Xcomb[2:end]...], Float64)
            if length(reterms) != length(Xcomb[2:end])
                throw("IncompatibleRandomGroupings")
            end
        catch e
            @error "Error, you seem to have two different eventtypes with the same random-effect grouping variable. \n
            This is not allowed, you have to rename one. Example:\n
            eventA: y~1+(1|item) \n
            eventB: y~1+(1|item)  \n
            This leads to this error. Rename the later one\n
            eventB: y~1+(1|itemB) "
            throw("IncompatibleRandomGroupings")
        end



    end

    if X1.formulas isa FormulaTerm
        # due to the assertion above, we can assume we have only 2 formulas here
        if X1.formulas.rhs isa Unfold.TimeExpandedTerm
            fcomb = Vector{FormulaTerm{<:InterceptTerm, <:TimeExpandedTerm}}(undef,2)
        else
            fcomb = Vector{FormulaTerm}(undef,2) # mass univariate case
        end
        fcomb[1] = X1.formulas
        fcomb[2] = X2.formulas
        return DesignMatrix(fcomb, Xcomb, [X1.events, X2.events])
    else
        if X1.formulas[1].rhs isa Unfold.TimeExpandedTerm
            # we can ignore length of X2, as it has to be a single formula due to the assertion above
            fcomb = Vector{FormulaTerm{<:InterceptTerm, <:TimeExpandedTerm}}(undef,length(X1.formulas)+1)
        else
            fcomb = Vector{FormulaTerm}(undef,length(X1.formulas)+1) # mass univariate case
        end
        fcomb[1:end-1] = X1.formulas
        fcomb[end] = X2.formulas
        return DesignMatrix(fcomb, Xcomb, [X1.events..., X2.events])
    end
end

function changeMatSize!(m, fe::AbstractSparseMatrix, remats)
    changeReMatSize!.(remats, Ref(m))
    fe = SparseMatrixCSC(m, fe.n, fe.colptr, fe.rowval, fe.nzval)
    return (fe, remats...)
end
function changeMatSize!(m, fe::AbstractMatrix, remats)
    changeReMatSize!.(remats, Ref(m))
    fe = fe[1:m,:]
    return (fe, remats...)
end
function changeReMatSize!(remat::MixedModels.AbstractReMat, m::Integer)

    n = m - length(remat.refs) # missing elements
    if n < 0
        deleteat!(remat.refs, range(m + 1, stop = length(remat.refs)))
        remat.adjA = remat.adjA[:, 1:(m+1)]
        remat.wtz = remat.wtz[:, 1:(m+1)]
        remat.z = remat.z[:, 1:(m+1)]
    elseif n > 0
        append!(remat.refs, repeat([remat.refs[end]], n))
        remat.adjA = [remat.adjA sparse(repeat([0], size(remat.adjA)[1], n))]
        remat.wtz = [remat.wtz zeros((size(remat.wtz)[1], n))]
        remat.z = [remat.z zeros((size(remat.z)[1], n))]
    end

    #ReMat{T,S}(remat.trm, refs, levels, cnames, z, wtz, λ, inds, adjA, scratch)
end
function equalizeReMatLengths!(remats::NTuple{A,MixedModels.AbstractReMat}) where {A}
    # find max length
    m = maximum([x[1] for x in size.(remats)])
    @debug print("combining lengths: $m")
    # for each reMat
    for k in range(1, length = length(remats))
        remat = remats[k]
        if size(remat)[1] == m
            continue
        end
        # prolong if necessary
        changeReMatSize!(remat, m)

    end
end


Base.:+(X1::DesignMatrix, X2::DesignMatrix) = combineDesignmatrices(X1, X2)

Base.:+(X1::Nothing, X2::DesignMatrix) = X2

function get_Xs(Xs::Tuple)
    return Xs[1]
end
function get_Xs(Xs::SparseMatrixCSC)
    return Xs
end
function get_Xs(Xs::Matrix)
    # mass univariate case
    return Xs
end
function get_Xs(Xs::Vector)
    # mass univariate case with multiple events
    return Xs
end
function get_Xs(X::DesignMatrix)
    return get_Xs(X.Xs)
end

"""
$(SIGNATURES)
designmatrix(type, f, tbl; kwargs...)
Return a *DesignMatrix* used to fit the models.
# Arguments
- type::UnfoldModel
- f::FormulaTerm: Formula to be used in this designmatrix
- tbl: Events (usually a data frame) to be modelled
- basisfunction::BasisFunction: basisfunction to be used in modeling (if specified)
- contrasts::Dict: (optional) contrast to be applied to formula
- eventfields::Array: (optional) Array of symbols which are passed to basisfunction event-wise. 
First field of array always defines eventonset in samples. Default is [:latency]

# Examples
```julia-repl
julia>  designmatrix(UnfoldLinearModelContinuousTime,Dict(Any=>(f,basisfunction1),tbl)
```

"""
function designmatrix(
    type,
    f::Union{Tuple,FormulaTerm},
    tbl,
    basisfunction;
    contrasts = Dict{Symbol,Any}(),
    kwargs...,
)
    @debug("generating DesignMatrix")

    # check for missings-columns - currently missings not really supported in StatsModels
    s_tmp = schema(f,tbl,contrasts) # temporary scheme to get necessary terms
    neededCols = [v.sym for v in values(s_tmp.schema)]

    tbl_nomissing = DataFrame(tbl) # tbl might be a SubDataFrame due to a view - but we can't modify a subdataframe, can we?
    try 
        disallowmissing!(tbl_nomissing,neededCols) # if this fails, it means we have a missing value in a column we need. We do not support this
    catch e
        if e isa ArgumentError
            error(e.msg*"\n we tried to get rid of a event-column declared as type Union{Missing,T}. But there seems to be some actual missing values in there. 
            You have to replace them yourself (e.g. replace(tbl.colWithMissingValue,missing=>0)) or impute them otherwise.")
        else
           rethrow()
        end
    end

        


    form = apply_schema(f, schema(f, tbl_nomissing, contrasts), MixedModels.LinearMixedModel)
    
    form =
        apply_basisfunction(form, basisfunction, get(Dict(kwargs), :eventfields, nothing))

    # Evaluate the designmatrix
    
    #note that we use tbl again, not tbl_nomissing.
    @debug typeof(form)
    if (!isnothing(basisfunction)) & (type <: UnfoldLinearMixedModel)
        X = modelcols.(form.rhs, Ref(tbl))
    else
        X = modelcols(form.rhs, tbl)
    end
    @debug typeof(X)
    return DesignMatrix(form, X, tbl)
end
function designmatrix(type, f, tbl; kwargs...)
    return designmatrix(type, f, tbl, nothing; kwargs...)
end

# specify for abstract interface
designmatrix(uf::UnfoldModel) = uf.designmatrix

function designmatrix(
    uf::UnfoldModel,
    tbl;
    eventcolumn = :event,
    contrasts = Dict{Symbol,Any}(),
    kwargs...,
)
    X = nothing
    fDict = design(uf)
    for (eventname, f) in pairs(fDict)
        
        @debug eventname,X
        if eventname == Any
            eventTbl = tbl
        else
            if !((eventcolumn ∈ names(tbl)) | (eventcolumn ∈ propertynames(tbl)))
                error(
                    "Couldnt find columnName: " *
                    string(eventcolumn) *
                    " in event-table.  Maybe need to specify eventcolumn=:correctColumnName (default is ':event') \n names(tbl) = " *
                    join(names(tbl), ","),
                )
            end
            eventTbl = @view tbl[tbl[:, eventcolumn].==eventname, :] # we need a view so we can remap later if needed
        end
        if isempty(eventTbl)
            error(
                "eventTable empty after subsetting. Couldnt find event '" *
                string(eventname) *
                "'{" *
                string(typeof(eventname)) *
                "}, in field tbl[:,:" *
                string(eventcolumn) *
                "].? - maybe you need to specify it as a string instead of a symbol?",
            )
        end

        fIx = collect(typeof.(f) .<: FormulaTerm)
        bIx = collect(typeof.(f) .<: BasisFunction)



        if any(bIx)
            # timeContinuos way
            # TODO there should be a julian way to do this distinction

            X =
                X + designmatrix(
                    typeof(uf),
                    f[fIx],
                    eventTbl,
                    collect(f[bIx])[1];
                    contrasts = contrasts,
                    kwargs...,
                )
        else
            # normal way
            X =
                X +
                designmatrix(typeof(uf), f[fIx], eventTbl; contrasts = contrasts, kwargs...)
        end
    end
    return X
end

import Base.isempty
Base.isempty(d::DesignMatrix) = isempty(d.Xs)

"""
$(SIGNATURES)
timeexpand the rhs-term of the formula with the basisfunction

"""
function apply_basisfunction(form, basisfunction::BasisFunction, eventfields)
    @debug("apply_basisfunction")
    return FormulaTerm(form.lhs, TimeExpandedTerm(form.rhs, basisfunction, eventfields))
end

function apply_basisfunction(form, basisfunction::Nothing, eventfields)
    # in case of no basisfunctin, do nothing
    return form
end


function designmatrix!(uf::UnfoldModel, evts; kwargs...)
    X = designmatrix(uf, evts; kwargs...)
    uf.designmatrix = X
end

function StatsModels.modelmatrix(uf::UnfoldLinearModel,basisfunction)
    if basisfunction
        @warn("basisfunction not defined for this kind of model")
    else
        return modelmatrix(uf)
    end
end

function StatsModels.modelmatrix(uf::UnfoldLinearModelContinuousTime,basisfunction = true)
    if basisfunction
        return modelmatrix(designmatrix(uf))
    else
        # replace basisfunction with non-timeexpanded one
        f = formula(uf)
        
        # probably a more julian way to do this...
        if isa(f,AbstractArray)
            return modelcols_nobasis.(f[1,:],events(uf))
        else
            return modelcols_nobasis(f,events(uf))
        end
        
    end
end

modelcols_nobasis(f::FormulaTerm,tbl::AbstractDataFrame) = modelcols(f.rhs.term,tbl)
StatsModels.modelmatrix(uf::UnfoldModel) = modelmatrix(designmatrix(uf))#modelmatrix(uf.design,uf.designmatrix.events)
StatsModels.modelmatrix(d::DesignMatrix) = d.Xs


StatsModels.modelmatrix(d::Dict, events) = modelcols(formula(d).rhs, events)

formula(uf::UnfoldModel) = formula(designmatrix(uf))
formula(d::DesignMatrix) = d.formulas


events(uf::UnfoldModel) = events(designmatrix(uf))
events(d::DesignMatrix) = d.events

design(uf::UnfoldModel) = uf.design

function formula(d::Dict) #TODO Specify Dict better
    if length(values(d)) == 1
        return [c[1] for c in collect(values(d))][1]
    else
        return [c[1] for c in collect(values(d))]
    end
    
end

"""
$(SIGNATURES)
calculates the actual designmatrix for a timeexpandedterm. Multiple dispatch on StatsModels.modelcols
"""
function StatsModels.modelcols(term::TimeExpandedTerm, tbl)
    @debug term.term , first(tbl)
    X = modelcols(term.term, tbl)

    time_expand(X, term, tbl)
end


mutable struct SparseReMat{T,S} <: MixedModels.AbstractReMat{T}
    trm::Any
    refs::Vector{Int32}
    levels::Any
    cnames::Vector{String}
    z::SparseMatrixCSC{T}
    wtz::SparseMatrixCSC{T}
    λ::LowerTriangular{T,Matrix{T}}
    inds::Vector{Int}
    adjA::SparseMatrixCSC{T,Int32}
    scratch::Matrix{T}
end

"""
$(SIGNATURES)
This function timeexpands the random effects and generates a ReMat object
"""
function StatsModels.modelcols(
    term::TimeExpandedTerm{<:Union{<:RandomEffectsTerm,<:MixedModels.AbstractReTerm}},
    tbl,
)
    # exchange this to get ZeroCorr to work


    tbl = DataFrame(tbl)
    # get the non-timeexpanded reMat
    reMat = modelcols(term.term, tbl)

    # Timeexpand the designmatrix
    z = transpose(time_expand(transpose(reMat.z), term, tbl))



    # First we check if there is overlap in the timeexpanded term. If so, we cannot continue. Later implementations will remedy that
    #println(dump(term,))
    if hasfield(typeof(term.term), :rhs)
        rhs = term.term.rhs


    elseif hasfield(typeof(term.term.term), :rhs)
        # we probably have something like zerocorr, which does not need to show a .rhs necessarily
        rhs = term.term.term.rhs
    else

        println("term.term: $(dump(term.term))")
        error("unknown RE structure, has no field .rhs:$(typeof(term.term))")

    end

    group = tbl[!, rhs.sym]
    time = tbl[!, term.eventfields[1]]

    # get the from-to onsets of the grouping varibales
    onsets = time_expand_getRandomGrouping(group, time, term.basisfunction)
    #print(size(reMat.z))
    refs = zeros(size(z)[2]) .+ 1
    for (i, o) in enumerate(onsets[2:end])
        # check for overlap
        if (minimum(o) <= maximum(onsets[i+1])) & (maximum(o) <= minimum(onsets[i+1]))
            error("overlap in random effects structure detected, not currently supported")
        end
    end

    # From now we can assume no overlap
    # We want to fnd which subject is active when
    refs = zeros(size(z)[2]) .+ 1
    uGroup = unique(group)

    for (i, g) in enumerate(uGroup[1:end])

        ix_start = findfirst(g .== group)
        ix_end = findlast(g .== group)
        if i == 1
            time_start = 1
        else
            time_start = time[ix_start]
            # XXX Replace this functionality by shiftOnset?
            time_start = time_start - sum(times(term.basisfunction) .<= 0)
        end
        if i == length(uGroup)
            time_stop = size(refs, 1)
        else
            time_stop = time[ix_end]
            time_stop = time_stop + sum(times(term.basisfunction) .> 0)
        end
        if time_start < 0
            time_start = 1
        end

        if time_stop > size(refs, 1)
            time_stop = size(refs, 1)
        end


        #println("$g,$time_start,$time_stop")
        refs[Int64(time_start):Int64(time_stop)] .= i
    end

    # Other variables with implementaions taken from the LinerMixedModel function
    wtz = z
    trm = term

    S = size(z, 1)
    T = eltype(z)
    λ = LowerTriangular(Matrix{T}(I, S, S))

    inds = MixedModels.sizehint!(Int[], (S * (S + 1)) >> 1) # double trouble? can this line go?
    m = reshape(1:abs2(S), (S, S))
    inds = sizehint!(Int[], (S * (S + 1)) >> 1)
    for j = 1:S
        for i = j:S
            # We currently restrict to diagonal entries
            # Once mixedmodels#293 is pushed, we can relax this and use zerocorr()
            if !(typeof(term.term) <: MixedModels.ZeroCorr) || (i == j) # for diagonal
                push!(inds, m[i, j])
            end
        end
    end

    levels = reMat.levels
    refs = refs


    # reMat.levels doesnt change
    cnames = coefnames(term)
    #print(refs)
    adjA = MixedModels.adjA(refs, z)
    scratch = Matrix{T}(undef, (S, length(uGroup)))

    # Once MixedModels.jl supports it, can be replaced with: SparseReMat
    ReMat{T,S}(rhs, refs, levels, cnames, z, wtz, λ, inds, adjA, scratch)
end

"""
$(SIGNATURES)
Get the timeranges where the random grouping variable was applied
"""
function time_expand_getRandomGrouping(tblGroup, tblLatencies, basisfunction)
    ranges = time_expand_getTimeRange.(tblLatencies, Ref(basisfunction))
end

# helper function to get the ranges from where to where the basisfunction is added
function time_expand_getTimeRange(onset, basisfunction)
    npos = sum(times(basisfunction) .>= 0)
    nneg = sum(times(basisfunction) .< 0)

    #basis = kernel(basisfunction)(onset)

    fromRowIx = floor(onset) - nneg
    toRowIx = floor(onset) + npos

    range(fromRowIx, stop = toRowIx)
end


"""
$(SIGNATURES)
performs the actual time-expansion in a sparse way.

 - Get the non-timeexpanded designmatrix X from StatsModels.
 - evaluate the basisfunction kernel at each event
 - calculate the necessary rows, cols and values for the sparse matrix
 Returns SparseMatrixCSC 
"""
function time_expand(X, term, tbl)
    ncolsBasis = size(kernel(term.basisfunction)(0), 2)
    X = reshape(X, size(X, 1), :)

    ncolsX = size(X)[2]
    nrowsX = size(X)[1]
    ncolsXdc = ncolsBasis * ncolsX

    # this is the predefined eventfield, usually "latency"
    #println(term.eventfields)
    #println(tbl[term.eventfields[1]][1:10])
    tbl = DataFrame(tbl)
    onsets = tbl[!, term.eventfields[1]]

    if typeof(term.eventfields) <: Array && length(term.eventfields) == 1
        bases = kernel(term.basisfunction).(tbl[!, term.eventfields[1]])
    else
        bases = kernel(term.basisfunction).(eachrow(tbl[!, term.eventfields]))
    end

    # generate rowindices
    rows = copy(rowvals.(bases))
    # this shift is necessary as some basisfunction time-points can be negative. But a matrix is always from 1:τ. Thus we have to shift it backwards in time.
    # The onsets are onsets-1 XXX not sure why.
    for r = 1:length(rows)
        rows[r] .+= floor(onsets[r] - 1) + shiftOnset(term.basisfunction)
    end

    rows = vcat(rows...)
    rows = repeat(rows, ncolsX)

    # generate column indices
    cols = []
    for Xcol = 1:ncolsX
        for b = 1:length(bases)
            for c = 1:ncolsBasis
                push!(
                    cols,
                    repeat([c + (Xcol - 1) * ncolsBasis], length(nzrange(bases[b], c))),
                )
            end
        end
    end
    cols = vcat(cols...)

    # generate values
    vals = []
    for Xcol = 1:ncolsX
        push!(vals, vcat(nonzeros.(bases) .* X[:, Xcol]...))
    end

    vals = vcat(vals...)
    ix = rows .> 0
    A = sparse(rows[ix], cols[ix], vals[ix])

    return A
end

"""
$(SIGNATURES)
coefnames of a TimeExpandedTerm concatenates the basis-function name with the kronecker product of the term name and the basis-function colnames. Separator is ' : '
Some examples for a firbasis:
        basis_313 : (Intercept) : 0.1
        basis_313 : (Intercept) : 0.2
        basis_313 : (Intercept) : 0.3
        ...
"""
function StatsModels.coefnames(term::TimeExpandedTerm)
    terms = coefnames(term.term)
    colnames = Unfold.colnames(term.basisfunction)
    name = Unfold.name(term.basisfunction)
    if typeof(terms) == String
        terms = [terms]
    end
    return name .* " : " .* kron(terms .* " : ", string.(colnames))
end

function termnames(term::TimeExpandedTerm)
    terms = coefnames(term.term)
    colnames = colnames(term.basisfunction)
    if typeof(terms) == String
        terms = [terms]
    end
    return vcat(repeat.([[t] for t in terms], length(colnames))...)
end


function colname_basis(term::TimeExpandedTerm)
    terms = coefnames(term.term)
    colnames = colnames(term.basisfunction)
    if typeof(terms) == String
        terms = [terms]
    end
    return repeat(colnames, length(terms))
end


function StatsModels.coefnames(terms::AbstractArray{<:FormulaTerm})
    return coefnames.(Base.getproperty.(terms, :rhs))
end
function StatsModels.coefnames(term::MixedModels.ZeroCorr)
    coefnames(term.term)
end

function StatsModels.coefnames(term::RandomEffectsTerm)
    coefnames(term.lhs)
end
